#!/usr/bin/php -q
<?php
// By Erik de Bruijn <reprap@erikdebruijn.nl>
// Description: this script will parse a G-Code file and reformat it to work with the newer RepRap G-Code interpreter firmware.
// Licence: GPL

$version = 0.4;
echo "\n( Modified by 3D-to-5D-Gcode v$version on ".date("c").')';
// Settings:

$setting['prepend_gcode'] = "";  // begin with this Gcode
$setting['append_gcode'] = "G91\nG1 Z4F100\nG90\nM104S200";  // add this Gcode at the end
$setting['extrusion_adjust'] = 1.5;  // factor to adjust extruded distances with
$setting['extrude_incremental'] = 1;  // E values are incremental. Better for making small adjustments
//$setting['replace_ereg'] = array('F330.0[^0-9]'=>'F1320.0');  // 
$setting['replace_strings'] = array(
	'F330.0'=>'F1320.0', // hugely speed up the raft-making
);  // 
$setting['remove_comments'] = true;  // 
$setting['soften_z_move_factor'] = 0.4;  // this slows down the move in the Z direction to this speed
$settings_file = $_SERVER['HOME']."/.reprap/3D-to-5D.settings";
if(file_exists($settings_file))
  $setting = ini_get($settings_file);
else 
  echo "\n( Warning: no settings file found in $settings_file using only defaults.)";
// explain defaults
foreach($setting as $sName => $sVal)
{
   $sVal = str_replace("\n"," \\n ",$sVal);
   echo "\n( $sName = $sVal )";
  if($aKey = array_search("--".$sName,$argv))
  {
    echo "\n( Cmdline override of setting: $sName ".$argv[($aKey+1)].')';
    $setting[$sName] = $argv[$aKey+1];
  }

}


// The *OLD*, 3D (lame!) way:
// M108S200 ; extruder motor speed at 200 PWM
// M101 ; extruder motor forward
// G1 ... ; make your moves...

// *NEW*, Now 5 Dimensionally!
// G1 X0.0 Y0.0 Z0.0
// G1 X1.0 Y1.0 E1.4142 ; Extrude the pythagorian distance (X^2+Y^2)^.5. Assuming just XY plain (2D) moves...
//
if($argv[1])
  $fName = $argv[1];
else 
  die("Usage: ".basename($argv[0])." foo.gcode [--setting-name setting-value]\n");

bcscale(4);
$lines = file($fName); // ugly mem-sucker... I know. When I have files that are too big I'll improve this ;) 
// if you fix it, please e-mail the update.

// Defaults: (check your firmware!)
$abs = false;
$X = $Y = $Z = "0.0";
$lastX = $lastY = $lastZ = "0.0";
$dist = 0;
$extruder_starting = false;

echo $setting['prepend_gcode'];
echo "\nG92 E0 \n";
//echo "\nG92 E0 ( reset extruder home pos)\n";

foreach($lines as $line)
{
  // Remove original comments
  if($setting['remove_comments']==true)
  { 
    if($line{0}==';') continue;
    if($line{0}=='\'') continue;
    if($line{0}=='(') continue;
  }
  #ereg("[GMEXYZFS]{1

  $comment = '';

  if(ereg("M[ ]*108[ ]S([0-9.]+)",$line,$regs))
  {
    $line .= " ; fw/bck/off";
    $speed = $regs[1];
    echo "(".trim($line).")\n";
    continue;
  }
  // Store previous coordinates
  $lastX = $X;
  $lastY = $Y;
  $lastZ = $Z;

  if(ereg("X[ ]*(-?[0-9.]+)",$line,$regs))
    $X = $regs[1];
  if(ereg("Y[ ]*(-?[0-9.]+)",$line,$regs))
    $Y = $regs[1];
  if(ereg("Z[ ]*(-?[0-9.]+)",$line,$regs))
    $Z = $regs[1];

  if(ereg("M[ ]*10([123])",$line,$regs))
  {
    switch($regs[1])
    {
      case "1":
        $dir = 1;
        $extruder_starting = true;
        // When starting extrusion, E = 0?
      break;
      case "2";
        $dir = -1;
      break;
      case "3":
        $dir = 0;
      break;
    }
    //$comment .= " dir $dir";
    //echo "(".trim($line).")\n";
    continue; // Assuming that M codes are always on a single line... skip this
  }
  if(ereg("G[ ]*90",$line))
    $abs = true;
  if(ereg("G[ ]*91",$line))
    $abs = false;
  // G20 - inches, G21 - mm
  unset($E);
  $comment .= sprintf("dXY=%.2f,%.2f ",$dX,$dY)." lastXY= ".sprintf("%.2f,%.2f",$lastX,$lastY);
  if($abs)
  {
/*    $dX=bcsub($X,$lastX);
    $dY=bcsub($Y,$lastY);
    $dZ=bcsub($Z,$lastZ);
*/
    $dX=(1.0*$X-$lastX);
    $dY=(1.0*$Y-$lastY);
    $dZ=(1.0*$Z-$lastZ);
 
  } else
  {
    $dX=$X;
    $dY=$Y;
    $dZ=$Z;
  }
    $dist2 = bcsqrt(bcadd(bcpow($dX,2),bcpow($dY,2)));
    if($setting['extrude_incremental'] != 1)
      $dist = bcadd($dist,$dist2);
    else
      $dist = $dist2;
    $comment .= "( dist=sqrt($dX^2+$dY^2)=".bcpow($dX,2).'+'.bcpow($dY,2).'= '.sprintf("%.2f",$dist).')';
  if($dir!=0) 
    $E = bcmul($setting['extrusion_adjust'],$dist);

  // Only when cartesian coordinates are used???
  //if(ereg("[^;(].*[XYZ])",$line,$regs))
  //{}

  $line = trim($line);
  if($extruder_starting)
  {
    $line .= " E0";
    $dist = 0;
    $extruder_starting = false;
  }
  if($dZ>0.1)
  {
    // Smooth-Z:
    if($setting['soften_z_move_factor'])
    {
      if(ereg("F([0-9\.]+)",$line,$regs))
      {
        $newF = bcmul($setting['soften_z_move_factor'],$regs[1]);
        if($setting['debug'])
          echo "(Smooth-Z dZ > 0.1 F=$regs[1], newF=$newF)\n";
        $line = ereg_replace("F[0-9\.]+","F$newF",$line);
      }
    }//end if smooth
  }
  elseif($E)
    $line .= " E".sprintf("%.2f",bcmul($E,$dir)); // number w/ four decimals

  // Replace feedrate:
  foreach($setting['replace_strings'] as $oldRate=>$newRate)
  {
    $line = str_replace("$oldRate","$newRate",$line);
  }
  
  echo $line;
  if($E > 999)
  {
    echo "\nG92 E0 ( set extruder home )";
    $dist = 0;
  }

 //echo " ; ".$comment;
  echo "\n";
  //

} // end foreach line

echo $setting['append_gcode'];
/*

$fp = fopen($fName,'r');
if(!$fp) die("File $argv[0] could not be opened.");

$go = true;
while($go)
{
if(!strpos($str,"\n"))
{
  $str .= fgets($fp,128);
}
else
{
  $lines = split("\n",$str);
  echo $str;
}  
}
*/
?>
